from kusto_tool.database import KustoDatabase
from kusto_tool.expression import TableExpr
from pytest import raises


def test_mv_expand():
    """mv-expand works"""
    tbl = TableExpr("tbl", KustoDatabase("c", "db"), columns=dict(foo=str))
    assert str(tbl.mv_expand(tbl.foo)) == "cluster('c').database('db').['tbl']\n| mv-expand foo\n"


def test_mv_expand_str():
    """mv-expand works"""
    tbl = TableExpr("tbl", KustoDatabase("c", "db"), columns=dict(foo=str))
    assert str(tbl.mv_expand("foo")) == "cluster('c').database('db').['tbl']\n| mv-expand foo\n"


def test_mv_expand_nonexistent_col():
    """mv-expand works"""
    tbl = TableExpr("tbl", KustoDatabase("c", "db"), columns=dict(foo=str))
    with raises(KeyError):
        tbl.mv_expand("bar")



"""Test fakes."""
from azure.kusto.data.response import KustoResponseDataSetV2, KustoResultTable


class FakeDatabase:
    """Fake database for testing."""

    def __init__(self, cluster, database):
        self.cluster = cluster
        self.database = database

    def execute(self, query):
        """Just return the query instead of running it."""
        return query


class FakeKustoResponseDataSet(KustoResponseDataSetV2):
    def __init__(self, tables):
        self.tables = tables
        self.query = None

    @property
    def primary_results(self):
        return self.tables


class FakeKustoResultTable(KustoResultTable):
    def __init__(self, columns, rows):
        self.columns = columns
        self.raw_rows = rows


class FakeKustoClient:
    """Fake KustoClient for testing."""

    def __init__(self, table):
        self.response = FakeKustoResponseDataSet([FakeKustoResultTable(table)])

    def execute_mgmt(self, database, query):
        self.response.query = query
        return self.response

    def execute_query(self, database, query):
        self.response.query = query
        return self.response



from datetime import datetime as dt
from datetime import timedelta as td
from decimal import Decimal
from typing import Any

from kusto_tool.expression import OP, Column, Prefix, typeof


def strcat(*args):
    """String concatenation.

    Parameters
    ----------
    args: list
        List of string Columns and/or scalar strings to concatenate.
    """
    return Prefix(OP.STRCAT, *args, dtype=str)


def sum(expr):
    """Sum a column or expression.

    Parameters
    ----------
    expr: str, Column or expression.
    """
    # if sum gets a string, it's referring to a Column in the TableExpr.
    if isinstance(expr, str):
        expr = Column(expr, float)
    return Prefix(OP.SUM, expr, agg=True, dtype=float)


def avg(expr):
    """Average a column or expression.

    Parameters
    ----------
    expr: str, Column or expression.
    """
    if isinstance(expr, str):
        expr = Column(expr, float)
    return Prefix(OP.AVG, expr, agg=True, dtype=float)


def mean(expr):
    """Average a column or expression.

    Parameters
    ----------
    expr: str, Column or expression.
    """
    return avg(expr)


def count():
    """Count rows in the result set."""
    return Prefix(OP.COUNT, agg=True, dtype=int)


def dcount(expr, accuracy=1):
    """Distinct count of a column.

    Parameters
    ----------
    expr: str, Column or expression.
        The column to apply distinct count to.
    accuracy: int, default 1
        The level of accuracy to apply to the hyper log log algorithm.
        Default is 1, the fastest but least accurate.
    """
    if isinstance(expr, str):
        expr = Column(expr, Any)
    return Prefix(OP.DCOUNT, expr, accuracy, agg=True, dtype=int)


def startofday(expr, offset=None):
    """Get the start of day for a timestamp (round to preceding midnight).

    Parameters
    ----------
    expr: str, Column or expression.
        The datetime column/expression to round to the start of the day.
    offset: int, default None
        The number of days to shift the date by (-1 subtracts a day, 1 adds a day.)
    """
    if offset:
        return Prefix("startofday", expr, offset, agg=False, dtype=dt)
    return Prefix("startofday", expr, agg=False, dtype=dt)


def endofday(expr, offset=None):
    """Get the end of day for a timestamp (round to 11:59:59.9999999).

    Parameters
    ----------
    expr: str, Column or expression.
        The datetime column/expression to round to the end of the day.
    offset: int, default None
        The number of days to shift the date by (-1 subtracts a day, 1 adds a day.)
    """
    if offset:
        return Prefix("endofday", expr, offset, agg=False, dtype=dt)
    return Prefix("endofday", expr, agg=False, dtype=dt)


def startofweek(expr, offset=None):
    """Get the start of the week for a timestamp (round to preceding Sunday at midnight).

    Parameters
    ----------
    expr: str, Column or expression.
        The datetime column/expression to round to the start of the week.
    offset: int, default None
        The number of weeks to shift the date by (-1 subtracts a week, 1 adds a week.)
    """
    if offset:
        return Prefix("startofweek", expr, offset, agg=False, dtype=dt)
    return Prefix("startofweek", expr, agg=False, dtype=dt)


def endofweek(expr, offset=None):
    """Get the end of the week for a timestamp (round to following Saturday at 11:59:59.9999999).

    Parameters
    ----------
    expr: str, Column or expression.
        The datetime column/expression to round to the start of the week.
    offset: int, default None
        The number of weeks to shift the date by (-1 subtracts a week, 1 adds a week.)
    """
    if offset:
        return Prefix("endofweek", expr, offset, agg=False, dtype=dt)
    return Prefix("endofweek", expr, agg=False, dtype=dt)


def startofmonth(expr, offset=None):
    """Get the start of the month for a timestamp (round to first of the month).

    Parameters
    ----------
    expr: str, Column or expression.
        The datetime column/expression to round to the start of the month.
    offset: int, default None
        The number of months to shift the date by (-1 subtracts a month, 1 adds a month.)
    """
    if offset:
        return Prefix("startofmonth", expr, offset, agg=False, dtype=dt)
    return Prefix("startofmonth", expr, agg=False, dtype=dt)


def endofmonth(expr, offset=None):
    """Get the end of the month for a timestamp (round to last day of month at 11:59:59.9999999).

    Parameters
    ----------
    expr: str, Column or expression.
        The datetime column/expression to round to the start of the month.
    offset: int, default None
        The number of months to shift the date by (-1 subtracts a month, 1 adds a month.)
    """
    if offset:
        return Prefix("endofmonth", expr, offset, agg=False, dtype=dt)
    return Prefix("endofmonth", expr, agg=False, dtype=dt)


def startofyear(expr, offset=None):
    """Get the start of the year for a timestamp (round to first of the year).

    Parameters
    ----------
    expr: str, Column or expression.
        The datetime column/expression to round to the start of the year.
    offset: int, default None
        The number of years to shift the date by (-1 subtracts a year, 1 adds a year.)
    """
    if offset:
        return Prefix("startofyear", expr, offset, agg=False, dtype=dt)
    return Prefix("startofyear", expr, agg=False, dtype=dt)


def endofyear(expr, offset=None):
    """Get the end of the year for a timestamp (round to last day of year at 11:59:59.9999999).

    Parameters
    ----------
    expr: str, Column or expression.
        The datetime column/expression to round to the start of the year.
    offset: int, default None
        The number of years to shift the date by (-1 subtracts a year, 1 adds a year.)
    """
    if offset:
        return Prefix("endofyear", expr, offset, agg=False, dtype=dt)
    return Prefix("endofyear", expr, agg=False, dtype=dt)


def tobool(expr):
    """Convert an expression to boolean."""
    return Prefix("tobool", expr, agg=False, dtype=bool)


def datetime(expr):
    """Construct a datetime literal."""
    return Prefix("datetime", expr, agg=False, dtype=dt)


def todatetime(expr):
    """Convert an expression to a datetime."""
    return Prefix("todatetime", expr, agg=False, dtype=dt)


def int_(expr):
    """Construct an int from a string literal."""
    return Prefix("int", expr, agg=False, dtype=int)


def toint(expr):
    """Converts input to an integer."""
    return Prefix("toint", expr, agg=False, dtype=int)


def long_(expr):
    """Construct a long from a string literal."""
    return Prefix("long", expr, agg=False, dtype=int)


def tolong(expr):
    """Convert an expression to long (signed 64 bit integer)."""
    return Prefix("tolong", expr, agg=False, dtype=int)


def decimal(expr):
    """Construct a decimal from a literal."""
    return Prefix("decimal", expr, agg=False, dtype=Decimal)


def todecimal(expr):
    """Convert an expression to decimal (fixed point)."""
    return Prefix("todecimal", expr, agg=False, dtype=Decimal)


def todouble(expr):
    """Convert an expression to double (signed 64 bit floating point)."""
    return Prefix("todouble", expr, agg=False, dtype=float)


def toreal(expr):
    """Convert an expression to real (signed 64 bit floating point)."""
    return Prefix("toreal", expr, agg=False, dtype=float)


def todynamic(expr):
    """Parse a JSON string to a dynamic."""
    return Prefix("todynamic", expr, agg=False, dtype=object)


def parse_json(expr):
    """Parse a JSON string to a dynamic."""
    return Prefix("parse_json", expr, agg=False, dtype=object)


def tohex(expr):
    """Converts input to a hexadecimal string."""
    return Prefix("tohex", expr, agg=False, dtype=str)


def tostring(expr):
    """Converts input to a string."""
    return Prefix("tostring", expr, agg=False, dtype=str)


def totimespan(expr):
    """Converts input to a timespan."""
    return Prefix("totimespan", expr, agg=False, dtype=td)


def time(expr):
    """Construct a timespan from a string literal."""
    return Prefix("time", expr, agg=False, dtype=td)


def ago(expr):
    """Subtract a timespan from the current time."""
    return Prefix("ago", expr, agg=False, dtype=dt)


# Math


def abs_(expr):
    """Absolute value."""
    return Prefix("abs", expr, agg=False, dtype=typeof(expr))


def sin(expr):
    """Sine function."""
    return Prefix("sin", expr, agg=False, dtype=float)


def asin(expr):
    """Arcsine function."""
    return Prefix("asin", expr, agg=False, dtype=float)


def cos(expr):
    """Cosine function."""
    return Prefix("cos", expr, agg=False, dtype=float)


def acos(expr):
    """Arccosine function."""
    return Prefix("acos", expr, agg=False, dtype=float)


def tan(expr):
    """Tangent function."""
    return Prefix("tan", expr, agg=False, dtype=float)


def atan(expr):
    """Arctangent function."""
    return Prefix("atan", expr, agg=False, dtype=float)


def cot(expr):
    """Cotangent function."""
    return Prefix("cot", expr, agg=False, dtype=float)


def exp(expr):
    """Exponentiation function e^x."""
    return Prefix("exp", expr, agg=False, dtype=float)


def exp2(expr):
    """Exponentiation function 2^x."""
    return Prefix("exp2", expr, agg=False, dtype=typeof(expr))


def exp10(expr):
    """Exponentiation function 10^x."""
    return Prefix("exp10", expr, agg=False, dtype=typeof(expr))


def floor(expr, round_to):
    """Rounds values down to an integer multiple of a given bin size."""
    return Prefix("floor", expr, round_to, agg=False, dtype=int)


def bin_(expr, round_to):
    """Rounds values down to an integer multiple of a given bin size."""
    return Prefix("bin", expr, round_to, agg=False, dtype=int)


def ceiling(expr):
    """Rounds up to the smallest integer greater than or equal to expr."""
    return Prefix("ceiling", expr, agg=False, dtype=int)


def gamma(expr):
    """Gamma function."""
    return Prefix("gamma", expr, agg=False, dtype=float)


def isinf(expr):
    """Returns True if expr is infinity (positive or negative)."""
    return Prefix("isinf", expr, agg=False, dtype=bool)


def isfinite(expr):
    """Returns False if expr is infinity (positive or negative)."""
    return Prefix("isfinite", expr, agg=False, dtype=bool)


def isnan(expr):
    """Returns True if expr is NaN (not a number)."""
    return Prefix("isnan", expr, agg=False, dtype=bool)


def jaccard_index(expr1, expr2):
    """Returns the Jaccard index (intersection over union) of two sets."""
    return Prefix("jaccard_index", expr1, expr2, agg=False, dtype=float)


def log(expr):
    """Returns the natural (base-e) logarithm of expr."""
    return Prefix("log", expr, agg=False, dtype=float)


def log2(expr):
    """Returns the binary (base-2) logarithm of expr."""
    return Prefix("log2", expr, agg=False, dtype=float)


def log10(expr):
    """Returns the decimal (base-10) logarithm of expr."""
    return Prefix("log10", expr, agg=False, dtype=float)


def loggamma(expr):
    """Returns the logarithm of the absolute value of the gamma function."""
    return Prefix("loggamma", expr, agg=False, dtype=float)


def sqrt(expr):
    """Square root function."""
    return Prefix("sqrt", expr, agg=False, dtype=float)


def sign(expr):
    """Returns -1 if expr is negative, 0 if zero, or 1 if positive."""
    return Prefix("sign", expr, agg=False, dtype=int)


def pi():
    """Returns the constant pi."""
    return Prefix("pi", agg=False, dtype=float)


def pow_(base, exponent):
    """Raises base to the exponent."""
    return Prefix("pow", base, exponent, agg=False, dtype=float)


def rand(maximum=1.0):
    """Return a random number between 0 and maximum (default 1.0)

    Parameters
    ----------
    maximum: float, default=1.0
        The maximum value for the randomly generated number.
    """
    return Prefix("rand", maximum, agg=False, dtype=float)



from kusto_tool import expression as exp
from pytest import fixture

from .fake_database import FakeDatabase


@fixture
def db():
    return FakeDatabase("test", "testdb")


@fixture
def tbl(db):
    return exp.TableExpr("tbl", database=db, columns={"foo": str, "bar": int})


def test_where_eq_str():
    actual = str(exp.Where(exp.Infix(exp.OP.EQ, exp.Column("foo", str), "a")))
    expected = "| where foo == 'a'"
    assert actual == expected


def test_where_eq_int():
    actual = str(exp.Where(exp.Infix(exp.OP.EQ, exp.Column("foo", str), 2)))
    expected = "| where foo == 2"
    assert actual == expected


def test_where_ne_int():
    actual = str(exp.Where(exp.Infix(exp.OP.NE, exp.Column("foo", str), 2)))
    expected = "| where foo != 2"
    assert actual == expected


def test_where_lt_int():
    actual = str(exp.Where(exp.Infix(exp.OP.LT, exp.Column("foo", str), 2)))
    expected = "| where foo < 2"
    assert actual == expected


def test_where_lte_int():
    actual = str(exp.Where(exp.Infix(exp.OP.LE, exp.Column("foo", str), 2)))
    expected = "| where foo <= 2"
    assert actual == expected


def test_where_gt_int():
    actual = str(exp.Where(exp.Infix(exp.OP.GT, exp.Column("foo", str), 2)))
    expected = "| where foo > 2"
    assert actual == expected


def test_where_gte_int():
    actual = str(exp.Where(exp.Infix(exp.OP.GE, exp.Column("foo", str), 2)))
    expected = "| where foo >= 2"
    assert actual == expected


def test_table_where_eq(tbl):
    q = str(tbl.where(tbl.bar == "foo"))
    ex = "cluster('test').database('testdb').['tbl']\n| where bar == 'foo'\n"
    assert q == ex


def test_table_where_ne(tbl):
    q = str(tbl.where(tbl.bar != "foo"))
    ex = "cluster('test').database('testdb').['tbl']\n| where bar != 'foo'\n"
    assert q == ex


def test_table_where_lt(tbl):
    q = str(tbl.where(tbl.bar < "foo"))
    ex = "cluster('test').database('testdb').['tbl']\n| where bar < 'foo'\n"
    assert q == ex


def test_table_where_le(tbl):
    q = str(tbl.where(tbl.bar <= "foo"))
    ex = "cluster('test').database('testdb').['tbl']\n| where bar <= 'foo'\n"
    assert q == ex


def test_table_where_gt(tbl):
    q = str(tbl.where(tbl.bar > "foo"))
    ex = "cluster('test').database('testdb').['tbl']\n| where bar > 'foo'\n"
    assert q == ex


def test_table_where_ge(tbl):
    q = str(tbl.where(tbl.bar >= "foo"))
    ex = "cluster('test').database('testdb').['tbl']\n| where bar >= 'foo'\n"
    assert q == ex


def test_where_repr():
    where = exp.Where(exp.Infix(exp.OP.EQ, exp.Column("foo", str), 2))
    assert repr(where) == "Where(Column(\"foo\", <class 'str'>) == 2)"


def test_where_and():
    foo = exp.Column("foo", str)
    bar = exp.Column("bar", str)
    where = exp.Where((foo == "a") & (bar == "b"))
    assert str(where) == "| where (foo == 'a') and (bar == 'b')"


def test_where_or():
    foo = exp.Column("foo", str)
    bar = exp.Column("bar", str)
    where = exp.Where((foo == "a") | (bar == "b"))
    assert str(where) == "| where (foo == 'a') or (bar == 'b')"


def test_not():
    foo = exp.Column("foo", bool)
    where = exp.Where(~(foo == "a"))
    assert str(where) == "| where not(foo == 'a')"


def test_where_cols():
    foo = exp.Column("foo", str)
    bar = exp.Column("bar", str)
    where = exp.Where(foo == bar)
    assert str(where) == "| where foo == bar"


def test_where_isin():
    foo = exp.Column("foo", str)
    where = exp.Where(foo.isin("bar", "baz"))
    assert str(where) == "| where foo in ('bar', 'baz')"


def test_where_isin_int():
    foo = exp.Column("foo", str)
    where = exp.Where(foo.isin(1, 2, 3))
    assert str(where) == "| where foo in (1, 2, 3)"



import kusto_tool.function as F
from kusto_tool.database import KustoDatabase
from kusto_tool.expression import Evaluate, TableExpr
from pytest import raises


def test_evaluate():
    """evaluate bag_unpack works"""
    tbl = TableExpr("tbl", KustoDatabase("cluster", "db"), columns={"foo": dict})
    expr = Evaluate(tbl.foo.bag_unpack())
    assert str(expr) == "| evaluate bag_unpack(foo)"


def test_evaluate_bag_unpack():
    """evaluate bag_unpack works when called on Column"""
    tbl = TableExpr("tbl", KustoDatabase("cluster", "db"), columns={"foo": "dynamic"})
    expr = tbl.evaluate(tbl.foo.bag_unpack())
    assert str(expr) == "cluster('cluster').database('db').['tbl']\n| evaluate bag_unpack(foo)\n"


def test_evaluate_bag_unpack_wrong_dtype():
    """evaluate bag_unpack errors on wrong data type"""
    tbl = TableExpr("tbl", KustoDatabase("cluster", "db"), columns={"foo": str})
    with raises(AssertionError):
        tbl.evaluate(tbl.foo.bag_unpack())



"""Classes for interacting with a Kusto database."""
import os
from collections.abc import KeysView
from pathlib import Path
from timeit import default_timer as timer

import jinja2 as jj
import pandas as pd
from azure.kusto.data import KustoClient, KustoConnectionStringBuilder
from azure.kusto.data.helpers import dataframe_from_result_table
from loguru import logger

from kusto_tool.expression import KTYPES, TableExpr, quote


def list_to_kusto(lst):
    """Convert a Python list to a Kusto list literal."""
    list_str = [quote(i) for i in list(lst)]
    return "dynamic([\n\t" + ",\n\t".join(list_str) + "\n])"


def dict_to_datatable(dictionary: dict) -> str:
    """Converts a dict to a Kusto datatable statement for use as a lookup table."""
    dict_str = "\n\t".join([f"'{k}', '{v}'," for k, v in dictionary.items()])
    template = """datatable(key: string, value: string)[
    {{ dict_str }}
]"""
    stmt = jj.Template(template).render(dict_str=dict_str)
    return stmt


def maybe_read_file(query):
    """Read contents from a file if the argument is a file path."""
    if os.path.isfile(query):
        logger.info("Reading file {}.", query)
        with open(query, "r", encoding="utf-8") as query_file:
            return query_file.read()
    return query


def render_template_query(query, *args, **kwargs) -> str:
    """Render a query with optional parameters."""
    # Any list arguments need to be converted to Kusto list strings
    query = maybe_read_file(query)
    converted_kwargs = {
        k: list_to_kusto(v) if isinstance(v, (list, tuple, set, KeysView)) else v
        for k, v in kwargs.items()
    }
    return jj.Template(query).render(*args, **converted_kwargs)


def render_set(query, table, folder, docstring, *args, replace=False, **kwargs) -> str:
    """Render a .set-or-[append|replace] command from a query."""
    query = maybe_read_file(query)
    query_rendered = render_template_query(query, *args, **kwargs)
    command = "replace" if replace else "append"
    set_append_template = """.set-or-{{ command }} {{ table }}
with (
folder = "{{ folder }}",
docstring = "{{ docstring }}",
)
<|
{{ query_string }}
"""
    command_rendered = render_template_query(
        set_append_template,
        command=command,
        table=table,
        folder=folder,
        docstring=docstring,
        query_string=query_rendered,
    )
    return command_rendered


class KustoDatabase:
    """A class representing a Kusto database."""

    def __init__(self, cluster, database, client=None):
        """A class representing a Kusto database.

        Parameters
        ----------
        cluster: str
            The cluster name.
        database: str
            The database name.
        client: KustoClient, default None
            Pass this if you wish to provide your own KustoClient.
        """
        self.cluster = cluster
        self.cluster_uri = f"https://{cluster}.kusto.windows.net"
        self.database = database
        kcsb = KustoConnectionStringBuilder.with_az_cli_authentication(self.cluster_uri)
        self.client = client or KustoClient(kcsb)

    def table(self, name, columns=None, inspect=False):
        """A tabular expression.

        Parameters
        ----------
        name: str
            The name of the table in the database.
        columns: dict or list
            Either:
            1. A dictionary where keys are column names and values are
            data type names, or
            2. A list of Column instances.
        inspect: bool, default False
            If true, columns will be inspected from the database. If columns
            list is provided and inspect is true, inspect takes precedence.

        Returns
        -------
        TableExpr
            A table expression instance.
        """
        if inspect:
            columns = self.execute(f".show table {name} cslschema").Schema
            if len(columns) < 1:
                raise KeyError(f"Table {name} does not exist in the database.")
            columns = columns.item()
            columns = columns.split(",")
            columns = {col.split(":")[0]: KTYPES[col.split(":")[1]] for col in columns}
        return TableExpr(name, database=self, columns=columns)

    def execute(self, query: str, *args, **kwargs):
        """Execute a query or command.

        Parameters
        ----------
        query: str
            The text of the Kusto query or command to run. Can also be a path to
            a file containing a query.
        args: List[Any]
            Positional arguments to pass to the query as Jinja2 template params.
        kwargs: Dict[Any]
            Keyword arguments to pass to the query as Jinja2 template params.

        Returns
        -------
        pandas.DataFrame
            A DataFrame containing the query results.
        """
        query = maybe_read_file(query)
        query_rendered = render_template_query(query, *args, **kwargs)

        method = (
            self.client.execute_mgmt
            if query_rendered.startswith(".")
            else self.client.execute_query
        )
        logger.info("Executing query on {}: {}", self.database, query_rendered)
        start_time = timer()
        result = method(self.database, query_rendered)
        end_time = timer()
        duration = end_time - start_time
        logger.info("Query execution completed in {:.2f} seconds.", duration)
        return dataframe_from_result_table(result.primary_results[0])

    def show_tables(self):
        """Show the list of tables in the database.

        Returns
        -------
        pandas.DataFrame
            A DataFrame listing all tables in the database, in the column "TableName".
        """
        return self.execute(".show tables")

    def drop_table(self, table):
        """Drop a table from the database.

        Parameters
        ----------
        table: str
            The name of the table to drop.

        Returns
        -------
        pandas.DataFrame
            A DataFrame containing the name of the dropped table, indicating success.
        """
        return self.execute(f".drop table {table}")

    def table_exists(self, table: str) -> bool:
        """Check if a table exists in the database.

        Parameters
        ----------
        table: str
            The name of the table to look for.

        Returns
        -------
        bool
            True if the table exists in the database.
        """
        tables = self.show_tables()
        return table in tables.TableName.tolist()

    def set_table(self, query, table, folder, docstring, *args, **kwargs):
        """
        Runs your query and appends or replaces the results to a table.

        Parameters
        ----------
        query: str
            The text of the Kusto query to run.
        table: str
            The table name to create.
        folder: str
            The kusto folder to save the table into.
        docstring: str
            The docstring for the table metadata.
        replace: bool, default False.
            Appends to the table if True, replaces contents if False.
        args: List[Any]
            Positional arguments to pass to the query as Jinja2 template params.
        kwargs: Dict[Any]
            Keyword arguments to pass to the query as Jinja2 template params.

        Returns
        -------
        pandas.DataFrame
            A DataFrame containing the results of the control command.
        """
        return self.execute(
            render_set(query, table, folder, docstring, replace=False, *args, **kwargs),
        )

    def to_parquet(self, query, path, *args, force=False, **kwargs):
        """Run the given query, cache the results as a local parquet file, and
        return results as a Pandas DataFrame.

        Parameters
        ----------
        query: str
            The text of the Kusto query to run.
        path: str
            The path to save the parquet file.
        force: bool, default False
            If False, the data will be read from the cached parquet file if it
            exists. If True, the data will be re-downloaded regardless.
        args: List[Any]
            Positional arguments to pass to the query as Jinja2 template params.
        kwargs: Dict[Any]
            Keyword arguments to pass to the query as Jinja2 template params.

        Returns
        -------
        pandas.DataFrame
            A DataFrame containing the results of the control command.
        """
        if not force:
            if os.path.isfile(path):
                logger.info("Reading dataframe from existing file {}", path)
                return pd.read_parquet(path)
        logger.info("File {} does not exist, will run query.", path)
        df = self.execute(query, *args, **kwargs)
        df.to_parquet(path, index=False)
        return df

    def __str__(self):
        return f"{str(self.cluster)}.database('{self.database}')"


class Cluster:
    """A class representing a Kusto cluster."""

    def __init__(self, name):
        self.name = name

    def database(self, name):
        """Create an instance representing a database in the cluster.

        Parameters
        ----------
        name: str
            The name of the Kusto database in the cluster.

        Returns
        -------
        KustoDatabase
            an instance representing the Kusto database.
        """
        return KustoDatabase(self.name, name)

    def __str__(self):
        return f"cluster('{self.name}')"


def cluster(name):
    """Convenience function to construct a Cluster instance.
    Makes the query look more like KQL.
    """
    return Cluster(name)



from datetime import datetime

import kusto_tool.function as F
from kusto_tool.expression import Column, Summarize, TableExpr

from .fake_database import FakeDatabase


def test_strcat():
    """strcat translates correctly"""
    assert str(F.strcat(Column("foo", str), "_", "bar")) == "strcat(foo, '_', 'bar')"


def test_dcount_str():
    """dcount works correctly for a string arg."""
    db = FakeDatabase("help", "Samples")
    tbl = TableExpr("tbl", database=db, columns={"foo": str, "bar": int})
    query = tbl.summarize(dcount_foo=F.dcount("foo", 2))
    assert "dcount_foo" in query.columns
    assert "bar" not in query.columns
    assert "foo" not in query.columns
    result = str(query)
    expected = """cluster('help').database('Samples').['tbl']
| summarize
\tdcount_foo=dcount(foo, 2)
"""
    assert result == expected


def test_dcount_col():
    """dcount works correctly for a column arg."""
    db = FakeDatabase("help", "Samples")
    tbl = TableExpr("tbl", database=db, columns={"foo": str, "bar": int})
    query = tbl.summarize(dcount_foo=tbl.foo.dcount(2))
    assert "dcount_foo" in query.columns
    assert "bar" not in query.columns
    assert "foo" not in query.columns
    result = str(query)
    expected = """cluster('help').database('Samples').['tbl']
| summarize
\tdcount_foo=dcount(foo, 2)
"""
    assert result == expected


def test_avg_col():
    """avg works correctly for a column arg."""
    db = FakeDatabase("help", "Samples")
    tbl = TableExpr("tbl", database=db, columns={"foo": str, "bar": int})
    query = tbl.summarize(avg_foo=tbl.foo.avg())
    assert "avg_foo" in query.columns
    assert "bar" not in query.columns
    assert "foo" not in query.columns
    result = str(query)
    expected = """cluster('help').database('Samples').['tbl']
| summarize
\tavg_foo=avg(foo)
"""
    assert result == expected


def test_avg_function_col():
    """avg works correctly for a column arg."""
    db = FakeDatabase("help", "Samples")
    tbl = TableExpr("tbl", database=db, columns={"foo": str, "bar": int})
    query = tbl.summarize(avg_foo=F.avg(tbl.foo))
    assert "avg_foo" in query.columns
    assert "bar" not in query.columns
    assert "foo" not in query.columns
    result = str(query)
    expected = """cluster('help').database('Samples').['tbl']
| summarize
\tavg_foo=avg(foo)
"""
    assert result == expected


def test_avg_function_str():
    """avg works correctly for a str arg."""
    db = FakeDatabase("help", "Samples")
    tbl = TableExpr("tbl", database=db, columns={"foo": str, "bar": int})
    query = tbl.summarize(avg_foo=F.avg("foo"))
    assert "avg_foo" in query.columns
    assert "bar" not in query.columns
    assert "foo" not in query.columns
    result = str(query)
    expected = """cluster('help').database('Samples').['tbl']
| summarize
\tavg_foo=avg(foo)
"""
    assert result == expected


def test_mean_function_col():
    """avg works correctly for a col arg."""
    db = FakeDatabase("help", "Samples")
    tbl = TableExpr("tbl", database=db, columns={"foo": str, "bar": int})
    query = tbl.summarize(avg_foo=F.mean(tbl.foo))
    assert "avg_foo" in query.columns
    assert "bar" not in query.columns
    assert "foo" not in query.columns
    result = str(query)
    expected = """cluster('help').database('Samples').['tbl']
| summarize
\tavg_foo=avg(foo)
"""
    assert result == expected


def test_mean_column():
    """avg works correctly."""
    db = FakeDatabase("help", "Samples")
    tbl = TableExpr("tbl", database=db, columns={"foo": str, "bar": int})
    query = tbl.summarize(avg_foo=tbl.foo.mean())
    assert "avg_foo" in query.columns
    assert "bar" not in query.columns
    assert "foo" not in query.columns
    result = str(query)
    expected = """cluster('help').database('Samples').['tbl']
| summarize
\tavg_foo=avg(foo)
"""
    assert result == expected


def test_count():
    """avg works correctly for a column arg."""
    db = FakeDatabase("help", "Samples")
    tbl = TableExpr("tbl", database=db, columns={"foo": str, "bar": int})
    query = tbl.summarize(ct=F.count())
    assert "ct" in query.columns
    assert "bar" not in query.columns
    assert "foo" not in query.columns
    result = str(query)
    expected = """cluster('help').database('Samples').['tbl']
| summarize
\tct=count()
"""
    assert result == expected


def test_startofday():
    """startofday works correctly for a column arg."""
    db = FakeDatabase("help", "Samples")
    tbl = TableExpr("tbl", database=db, columns={"foo": datetime})
    query = tbl.project(daystart=F.startofday(tbl.foo))
    assert "daystart" in query.columns
    assert "foo" not in query.columns
    result = str(query)
    expected = """cluster('help').database('Samples').['tbl']
| project
\tdaystart = startofday(foo)
"""
    assert result == expected


def test_startofday_offset():
    """startofday works correctly for a column arg."""
    db = FakeDatabase("help", "Samples")
    tbl = TableExpr("tbl", database=db, columns={"foo": datetime})
    query = tbl.project(daystart=F.startofday(tbl.foo, 1))
    assert "daystart" in query.columns
    assert "foo" not in query.columns
    result = str(query)
    expected = """cluster('help').database('Samples').['tbl']
| project
\tdaystart = startofday(foo, 1)
"""
    assert result == expected


def test_endofday():
    """startofday works correctly for a column arg."""
    db = FakeDatabase("help", "Samples")
    tbl = TableExpr("tbl", database=db, columns={"foo": datetime})
    query = tbl.project(daystart=F.endofday(tbl.foo))
    assert "daystart" in query.columns
    assert "foo" not in query.columns
    result = str(query)
    expected = """cluster('help').database('Samples').['tbl']
| project
\tdaystart = endofday(foo)
"""
    assert result == expected


def test_endofday_offset():
    """startofday works correctly for a column arg."""
    db = FakeDatabase("help", "Samples")
    tbl = TableExpr("tbl", database=db, columns={"foo": datetime})
    query = tbl.project(daystart=F.endofday(tbl.foo, 1))
    assert "daystart" in query.columns
    assert "foo" not in query.columns
    result = str(query)
    expected = """cluster('help').database('Samples').['tbl']
| project
\tdaystart = endofday(foo, 1)
"""
    assert result == expected


def test_startofweek():
    """startofweek works correctly for a column arg."""
    db = FakeDatabase("help", "Samples")
    tbl = TableExpr("tbl", database=db, columns={"foo": datetime})
    query = tbl.project(weekstart=F.startofweek(tbl.foo))
    assert "weekstart" in query.columns
    assert "foo" not in query.columns
    result = str(query)
    expected = """cluster('help').database('Samples').['tbl']
| project
\tweekstart = startofweek(foo)
"""
    assert result == expected


def test_startofweek_offset():
    """startofweek works correctly for a column arg."""
    db = FakeDatabase("help", "Samples")
    tbl = TableExpr("tbl", database=db, columns={"foo": datetime})
    query = tbl.project(weekstart=F.startofweek(tbl.foo, 1))
    assert "weekstart" in query.columns
    assert "foo" not in query.columns
    result = str(query)
    expected = """cluster('help').database('Samples').['tbl']
| project
\tweekstart = startofweek(foo, 1)
"""
    assert result == expected


def test_endofweek():
    """endofweek works correctly for a column arg."""
    db = FakeDatabase("help", "Samples")
    tbl = TableExpr("tbl", database=db, columns={"foo": datetime})
    query = tbl.project(weekend=F.endofweek(tbl.foo))
    assert "weekend" in query.columns
    assert "foo" not in query.columns
    result = str(query)
    expected = """cluster('help').database('Samples').['tbl']
| project
\tweekend = endofweek(foo)
"""
    assert result == expected


def test_endofweek_offset():
    """endofweek works correctly for a column arg."""
    db = FakeDatabase("help", "Samples")
    tbl = TableExpr("tbl", database=db, columns={"foo": datetime})
    query = tbl.project(weekend=F.endofweek(tbl.foo, 1))
    assert "weekend" in query.columns
    assert "foo" not in query.columns
    result = str(query)
    expected = """cluster('help').database('Samples').['tbl']
| project
\tweekend = endofweek(foo, 1)
"""
    assert result == expected


def test_startofmonth():
    """startofmonth works correctly for a column arg."""
    db = FakeDatabase("help", "Samples")
    tbl = TableExpr("tbl", database=db, columns={"foo": datetime})
    query = tbl.project(monthstart=F.startofmonth(tbl.foo))
    assert "monthstart" in query.columns
    assert "foo" not in query.columns
    result = str(query)
    expected = """cluster('help').database('Samples').['tbl']
| project
\tmonthstart = startofmonth(foo)
"""
    assert result == expected


def test_startofmonth_offset():
    """startofmonth works correctly for a column arg."""
    db = FakeDatabase("help", "Samples")
    tbl = TableExpr("tbl", database=db, columns={"foo": datetime})
    query = tbl.project(monthstart=F.startofmonth(tbl.foo, 1))
    assert "monthstart" in query.columns
    assert "foo" not in query.columns
    result = str(query)
    expected = """cluster('help').database('Samples').['tbl']
| project
\tmonthstart = startofmonth(foo, 1)
"""
    assert result == expected


def test_endofmonth():
    """endofmonth works correctly for a column arg."""
    db = FakeDatabase("help", "Samples")
    tbl = TableExpr("tbl", database=db, columns={"foo": datetime})
    query = tbl.project(monthend=F.endofmonth(tbl.foo))
    assert "monthend" in query.columns
    assert "foo" not in query.columns
    result = str(query)
    expected = """cluster('help').database('Samples').['tbl']
| project
\tmonthend = endofmonth(foo)
"""
    assert result == expected


def test_endofmonth_offset():
    """endofmonth works correctly for a column arg."""
    db = FakeDatabase("help", "Samples")
    tbl = TableExpr("tbl", database=db, columns={"foo": datetime})
    query = tbl.project(monthend=F.endofmonth(tbl.foo, 1))
    assert "monthend" in query.columns
    assert "foo" not in query.columns
    result = str(query)
    expected = """cluster('help').database('Samples').['tbl']
| project
\tmonthend = endofmonth(foo, 1)
"""
    assert result == expected


def test_startofyear():
    """startofyear works correctly for a column arg."""
    db = FakeDatabase("help", "Samples")
    tbl = TableExpr("tbl", database=db, columns={"foo": datetime})
    query = tbl.project(yearstart=F.startofyear(tbl.foo))
    assert "yearstart" in query.columns
    assert "foo" not in query.columns
    result = str(query)
    expected = """cluster('help').database('Samples').['tbl']
| project
\tyearstart = startofyear(foo)
"""
    assert result == expected


def test_startofyear_offset():
    """startofyear works correctly for a column arg."""
    db = FakeDatabase("help", "Samples")
    tbl = TableExpr("tbl", database=db, columns={"foo": datetime})
    query = tbl.project(yearstart=F.startofyear(tbl.foo, 1))
    assert "yearstart" in query.columns
    assert "foo" not in query.columns
    result = str(query)
    expected = """cluster('help').database('Samples').['tbl']
| project
\tyearstart = startofyear(foo, 1)
"""
    assert result == expected


def test_endofyear():
    """endofyear works correctly for a column arg."""
    db = FakeDatabase("help", "Samples")
    tbl = TableExpr("tbl", database=db, columns={"foo": datetime})
    query = tbl.project(yearend=F.endofyear(tbl.foo))
    assert "yearend" in query.columns
    assert "foo" not in query.columns
    result = str(query)
    expected = """cluster('help').database('Samples').['tbl']
| project
\tyearend = endofyear(foo)
"""
    assert result == expected


def test_endofyear_offset():
    """endofyear works correctly for a column arg."""
    db = FakeDatabase("help", "Samples")
    tbl = TableExpr("tbl", database=db, columns={"foo": datetime})
    query = tbl.project(yearend=F.endofyear(tbl.foo, 1))
    assert "yearend" in query.columns
    assert "foo" not in query.columns
    result = str(query)
    expected = """cluster('help').database('Samples').['tbl']
| project
\tyearend = endofyear(foo, 1)
"""
    assert result == expected



from kusto_tool.expression import OP, Column, Extend, Infix, Prefix, typeof


def test_extend():
    """extend renders"""
    assert str(Extend(foo="bar")) == "| extend\n\tfoo='bar'"


def test_extend_two():
    """extend renders with two kwargs"""
    assert str(Extend(foo="bar", baz="quux")) == "| extend\n\tfoo='bar',\n\tbaz='quux'"


def test_typeof_bin():
    """type of binary expression works"""
    assert typeof(Column("foo", int) + 1) == int


def test_typeof_unary():
    """type of unary expression works"""
    assert typeof(~Column("foo", int)) == bool


def test_typeof():
    """type of eagerly evaluated Python expression works"""
    assert typeof("foo" + "bar") == str


def test_typeof_add_int_int():
    assert typeof(Column("foo", int) + Column("bar", int)) == int


def test_typeof_add_int_float():
    assert typeof(Column("foo", int) + Column("bar", float)) == float


def test_typeof_add_float_int():
    assert typeof(Column("foo", float) + Column("bar", int)) == float


def test_typeof_sub_float_int():
    assert typeof(Column("foo", float) - Column("bar", int)) == float


def test_typeof_sub_int_int():
    assert typeof(Column("foo", int) - Column("bar", int)) == int


def test_typeof_div_float_int():
    """type of eagerly evaluated Python expression works"""
    assert typeof(Column("foo", float) / Column("bar", int)) == float


def test_typeof_mul_float_int():
    """type of eagerly evaluated Python expression works"""
    assert typeof(Column("foo", float) * Column("bar", int)) == float


def test_typeof_mul_int_int():
    """type of eagerly evaluated Python expression works"""
    assert typeof(Column("foo", int) * Column("bar", int)) == int



from kusto_tool.expression import Join, TableExpr

from .fake_database import FakeDatabase


def test_join_str():
    t2 = TableExpr("table2", "db", columns=dict(foo=str, bar=str, baz=int))
    join = Join(t2, on=["foo", "bar"], kind="inner")
    expected = """| join kind=inner hint.strategy=shuffle (
    table2) on foo, bar"""
    assert str(join == expected)


def test_table_shuffle_join():
    db = FakeDatabase("test", "testdb")
    t1 = TableExpr("table1", db, columns=dict(foo=str, bar=str, baz=int))
    t2 = TableExpr("table2", db, columns=dict(foo=str, bar=str, baz=int))
    join = t1.join(t2, on=["foo", "bar"], kind="inner", strategy="shuffle")
    expected = """cluster('test').database('testdb').['table1']
| join kind=inner hint.strategy=shuffle (
\tcluster('test').database('testdb').['table2']
) on foo, bar
"""
    print(expected)
    print(join)
    assert str(join) == expected


def test_table_shuffle_join():
    db = FakeDatabase("test", "testdb")
    t1 = TableExpr("table1", db, columns=dict(foo=str, bar=str, baz=int))
    t2 = TableExpr("table2", db, columns=dict(foo=str, bar=str, baz=int))
    join = t1.join(t2, on=["foo", "bar"], kind="inner")
    expected = """cluster('test').database('testdb').['table1']
| join kind=inner (
\tcluster('test').database('testdb').['table2']
) on foo, bar
"""
    print(expected)
    print(join)
    assert str(join) == expected


def test_broadcast_join_str():
    db = FakeDatabase("test", "testdb")
    t1 = TableExpr("table1", db, columns=dict(foo=str, bar=str, baz=int))
    result = str(Join(t1, on="foo", kind="inner", strategy="broadcast"))
    expected = """| join kind=inner hint.strategy=broadcast (
\tcluster('test').database('testdb').['table1']
) on foo"""
    assert result == expected


def test_other_join_str():
    db = FakeDatabase("test", "testdb")
    t1 = TableExpr("table1", db, columns=dict(foo=str, bar=str, baz=int))
    result = str(Join(t1, on="foo", kind="inner", strategy="silly"))
    expected = """| join kind=inner (
\tcluster('test').database('testdb').['table1']
) on foo"""
    assert result == expected



from azure.kusto.data.helpers import dataframe_from_result_table
from pytest import raises

from kusto_tool import database as kdb
from kusto_tool import expression as exp

from .fake_database import FakeDatabase, FakeKustoClient, FakeKustoResultTable


def test_dict_to_datatable():
    dct = {"foo": "one", "bar": "two", "baz": "three"}
    expected = (
        "datatable(key: string, value: string)[\n"
        "    'foo', 'one',\n\t'bar', 'two',\n\t'baz', 'three',\n]"
    )
    assert kdb.dict_to_datatable(dct) == expected


def test_list_to_kusto():
    """A list of strings renders as a Kusto list."""
    lst = ["foo", "bar", "baz"]
    expected = "dynamic([\n\t'foo',\n\t'bar',\n\t'baz'\n])"
    assert kdb.list_to_kusto(lst) == expected


def test_list_to_kusto_int():
    """A list of ints renders as a Kusto list."""
    lst = [1, 2, 3]
    expected = "dynamic([\n\t1,\n\t2,\n\t3\n])"
    assert kdb.list_to_kusto(lst) == expected


def test_render_template_query():
    foo = "bar"
    actual = kdb.render_template_query("{{ foo }} | take 10", foo=foo)
    expected = "bar | take 10"
    assert actual == expected


def test_render_set():
    result = kdb.render_set(
        "StormEvents | take 10",
        table="StormEventsTake10",
        folder="myfolder",
        docstring="mydocstring",
    )
    expected = """.set-or-append StormEventsTake10
with (
folder = "myfolder",
docstring = "mydocstring",
)
<|
StormEvents | take 10"""
    assert result == expected


def test_render_set_replace():
    result = kdb.render_set(
        "StormEvents | take 10",
        table="StormEventsTake10",
        folder="myfolder",
        docstring="mydocstring",
        replace=True,
    )
    expected = """.set-or-replace StormEventsTake10
with (
folder = "myfolder",
docstring = "mydocstring",
)
<|
StormEvents | take 10"""
    assert result == expected


def test_tableexpr_getattr():
    """Columns can be got from table with . operator"""
    db = kdb.KustoDatabase("test", "testdb")
    tbl = db.table("tbl", columns={"foo": str, "bar": int})
    assert tbl.foo.dtype == str
    assert tbl.bar.dtype == int


def test_tableexpr_getattr_bracket():
    """Columns can be got from table with [] operator"""
    db = kdb.KustoDatabase("test", "testdb")
    tbl = db.table("tbl", columns={"foo": str, "bar": int})
    assert tbl["foo"].dtype == str
    assert tbl["bar"].dtype == int


def test_tableexpr_getattr_column():
    """Creating table with list of Column works."""
    db = kdb.KustoDatabase("test", "testdb")
    tbl = db.table("tbl", columns=[exp.Column("foo", str), exp.Column("bar", int)])
    assert tbl.foo.dtype == str
    assert tbl.bar.dtype == int


def test_unknown_column_raises():
    """Accessing an unknown column errors."""
    db = kdb.KustoDatabase("test", "testdb")
    tbl = db.table("tbl", columns=[exp.Column("foo", str), exp.Column("bar", int)])
    with raises(AttributeError):
        return tbl.baz


def test_no_columns():
    """A table can be created with no columns, but accessing one errors."""
    db = kdb.KustoDatabase("test", "testdb")
    tbl = db.table("tbl")
    with raises(AttributeError):
        return tbl.baz


def test_columns_valueerror():
    """Passing something else as columns errors."""
    db = kdb.KustoDatabase("test", "testdb")
    with raises(ValueError):
        return db.table("tbl", columns=db)


def test_tableexpr_project():
    """Project statement generates correctly."""
    db = kdb.KustoDatabase("test", "testdb")
    tbl = db.table("tbl", columns={"foo": str, "bar": int})
    query = str(tbl.project(tbl.foo, tbl.bar, baz=tbl.bar))
    expected = (
        "cluster('test').database('testdb').['tbl']\n| project\n\tfoo,\n\tbar,\n\tbaz = bar\n"
    )
    assert query == expected


def test_collect():
    db = FakeDatabase("test", "testdb")
    tbl = kdb.TableExpr("tbl", database=db, columns={"foo": str, "bar": int})
    query = tbl.project(tbl.foo, tbl.bar, baz=tbl.bar).collect()
    expected = (
        "cluster('test').database('testdb').['tbl']\n| project\n\tfoo,\n\tbar,\n\tbaz = bar\n"
    )
    assert query == expected


def test_count():
    db = FakeDatabase("test", "testdb")
    tbl = kdb.TableExpr("tbl", database=db, columns={"foo": str, "bar": int})
    query = str(tbl.project(tbl.foo, tbl.bar).count())
    expected = "cluster('test').database('testdb').['tbl']\n| project\n\tfoo,\n\tbar\n| count\n"
    assert query == expected


def test_count_repr():
    assert repr(exp.Count()) == "Count()"


def test_table_distinct():
    db = FakeDatabase("test", "testdb")
    tbl = kdb.TableExpr("tbl", database=db, columns={"foo": str, "bar": int})
    q = str(tbl.distinct(tbl.bar, tbl.foo))
    ex = "cluster('test').database('testdb').['tbl']\n| distinct bar,\nfoo\n"
    assert q == ex


def test_distinct():
    query = str(exp.Distinct(exp.Column("bar", str), exp.Column("foo", int)))
    expected = "| distinct bar,\nfoo"
    assert query == expected


def test_column_repr():
    assert repr(exp.Distinct("foo", "bar")) == "Distinct(foo, bar)"


def test_expression_repr():
    assert repr(exp.Column("foo", str)) == "Column(\"foo\", <class 'str'>)"


def test_get_columns():
    db = FakeDatabase("test", "testdb")
    tbl = kdb.TableExpr("tbl", database=db, columns={"foo": str})
    assert tbl.foo == tbl.columns["foo"]


def test_extend():
    db = FakeDatabase("test", "testdb")
    tbl = kdb.TableExpr("tbl", database=db, columns={"foo": str})
    tbl_2 = tbl.extend(bar="baz")
    query = str(tbl_2)
    expected = "cluster('test').database('testdb').['tbl']\n| extend\n\tbar='baz'\n"
    assert "bar" in tbl_2.columns
    assert "bar" not in tbl.columns
    assert query == expected


def test_summarize_adds_aliases():
    """summarize adds aliases to column list."""
    tbl = (
        kdb.cluster("help")
        .database("Samples")
        .table("StormEvents", columns={"State": str, "EventType": str, "DamageProperty": int})
    )
    query = tbl.project(tbl.State, tbl.EventType, tbl.DamageProperty).summarize(
        sum_damage=tbl.DamageProperty.sum(), by=[tbl.State, tbl.EventType]
    )
    query = query.sort(query.sum_damage).limit(20)
    expected = """cluster('help').database('Samples').['StormEvents']
| project
\tState,
\tEventType,
\tDamageProperty
| summarize
\tsum_damage=sum(DamageProperty)
\tby State, EventType
| order by
\tsum_damage
| limit 20
"""
    assert str(query) == expected


def test_long_query():
    """test a longer query"""
    tbl = (
        kdb.cluster("help")
        .database("Samples")
        .table("StormEvents", columns={"State": str, "EventType": str, "DamageProperty": int})
    )
    query = (
        tbl.project(tbl.State, tbl.EventType, tbl.DamageProperty)
        .summarize(sum_damage=tbl.DamageProperty.sum(), by=[tbl.State, tbl.EventType])
        .sort("sum_damage")
        .limit(20)
    )
    expected = """cluster('help').database('Samples').['StormEvents']
| project
\tState,
\tEventType,
\tDamageProperty
| summarize
\tsum_damage=sum(DamageProperty)
\tby State, EventType
| order by
\tsum_damage
| limit 20
"""
    assert str(query) == expected


def test_expr_idempotent():
    """Querying a tbl multiple times is idempotent."""
    tbl = (
        kdb.cluster("help")
        .database("Samples")
        .table("StormEvents", columns={"State": str, "EventType": str, "DamageProperty": int})
    )
    query = tbl.project(tbl.State, tbl.EventType, tbl.DamageProperty)
    query = tbl.project(tbl.State, tbl.EventType, tbl.DamageProperty)
    expected = """cluster('help').database('Samples').['StormEvents']
| project
\tState,
\tEventType,
\tDamageProperty
"""
    assert str(query) == expected



from kusto_tool.database import KustoDatabase
from kusto_tool.expression import Column, Property, TableExpr
from pytest import raises


def test_project_rename_col():
    tbl = TableExpr("tbl", KustoDatabase("test", "testdb"), columns={"foo": int})
    expr = tbl.project(bar=tbl.foo)
    assert expr.columns["bar"].dtype == int
    assert "foo" not in expr.columns
    query = str(expr)
    expected = "cluster('test').database('testdb').['tbl']\n| project\n\tbar = foo\n"
    assert query == expected


def test_project_math_expr():
    tbl = TableExpr("tbl", KustoDatabase("test", "testdb"), columns={"foo": int})
    expr = tbl.project(bar=tbl.foo + 1)
    assert expr.columns["bar"].dtype == int
    assert "foo" not in expr.columns
    query = str(expr)
    expected = "cluster('test').database('testdb').['tbl']\n| project\n\tbar = foo + 1\n"
    assert query == expected


def test_property_access_table():
    """Property access succeeds for dict"""
    tbl = TableExpr("tbl", KustoDatabase("test", "testdb"), columns={"foo": dict})
    expr = tbl.project(bar=tbl.foo.baz)
    query = str(expr)
    assert query == "cluster('test').database('testdb').['tbl']\n| project\n\tbar = foo.baz\n"


def test_property_access():
    """Property access succeeds for dict"""
    assert str(Property(Column("foo", dict), "bar")) == "foo.bar"


def test_property_access_wrong_dtype():
    """Property access errors for non-dynamic column"""
    tbl = TableExpr("tbl", KustoDatabase("test", "testdb"), columns={"foo": str})
    with raises(AttributeError):
        expr = tbl.project(bar=tbl.foo.baz)



"""Experimental Kusto expression API for generating queries."""

from copy import copy
from datetime import datetime, timedelta
from decimal import Decimal
from typing import Any


class attrdict:
    """A dict whose members are accessible with the . operator."""

    def __init__(self, **kwargs):
        self._dict = kwargs

    def __getattr__(self, key):
        return self._dict[key]


OP = attrdict(
    CONTAINS="contains",
    EQ="==",
    GE=">=",
    GT=">",
    HAS="has",
    IN="in",
    LE="<=",
    LT="<",
    NCONTAINS="!contains",
    NE="!=",
    NHAS="!has",
    NIN="!in",
    SUM="sum",
    MIN="min",
    MAX="max",
    AND="and",
    OR="or",
    NOT="not",
    ADD="+",
    SUB="-",
    MUL="*",
    DIV="/",
    COUNT="count",
    DCOUNT="dcount",
    AVG="avg",
    STRCAT="strcat",
    BAG_UNPACK="bag_unpack",
    PERCENTILE="percentile",
)

PTYPES = {
    bool: "bool",
    datetime: "datetime",
    Decimal: "decimal",
    dict: "dynamic",
    float: "real",
    int: "long",
    list: "dynamic",
    object: "dynamic",
    str: "string",
    timedelta: "timespan",
    tuple: "dynamic",
}

KTYPES = {
    "bool": bool,
    "datetime": datetime,
    "decimal": Decimal,
    "dynamic": object,
    "int": int,
    "long": int,
    "real": float,
    "string": str,
    "timespan": timedelta,
}


def quote(val):
    """Quote strings."""
    if isinstance(val, str):
        return f"'{val}'"
    return str(val)


class Prefix:
    def __init__(self, op, *args, agg=False, dtype=Any):
        self.terms = args
        self.op = op
        self.agg = agg
        self.dtype = dtype

    def __str__(self):
        terms = ", ".join([quote(term) for term in self.terms])
        return f"{self.op}({terms})"


class Between:
    def __init__(self, lhs, left, right, negate=False):
        self.lhs = lhs
        self.left = left
        self.right = right
        self.negate = negate
        self.dtype = bool

    def __str__(self):
        neg = "!" if self.negate else ""
        return f"{self.lhs} {neg}between({self.left} .. {self.right})"


class Infix:
    # TODO: does this need __add__, __sub__?
    def __init__(self, op, lhs, rhs, dtype=Any):
        self.op = op
        self.lhs = lhs
        self.rhs = rhs
        self.dtype = dtype

    def __str__(self):
        if self.op in [OP.AND, OP.OR]:
            return f"({self.lhs}) {self.op} ({quote(self.rhs)})"
        return f"{self.lhs} {self.op} {quote(self.rhs)}"

    def __repr__(self):
        return f"{repr(self.lhs)} {self.op} {quote(self.rhs)}"

    def __and__(self, rhs):
        return Infix(OP.AND, self, rhs, dtype=bool)

    def __or__(self, rhs):
        return Infix(OP.OR, self, rhs, dtype=bool)

    def __invert__(self):
        return Prefix(OP.NOT, self, dtype=bool)


def typeof(expr):
    """Get the type of an expression."""
    if isinstance(expr, (Infix, Prefix, Column)):
        return expr.dtype
    return type(expr)


class Project:
    def __init__(self, *args, **kwargs):
        self.columns = list(args)
        self.renamed_columns = kwargs

    def _build_column_list(self):
        col_list = self.columns
        for k, v in self.renamed_columns.items():
            col_list.append(f"{k} = {v}")
        col_str = ",\n\t".join([str(col) for col in col_list])
        return col_str

    def __str__(self):
        column_list = self._build_column_list()
        return f"| project\n\t{column_list}"


class Count:
    def __repr__(self):
        return "Count()"

    def __str__(self):
        return "| count"


class Sample:
    """Sample operator."""

    def __init__(self, n):
        self.n = n

    def __repr__(self):
        return f"Sample({self.n})"

    def __str__(self):
        return f"| sample {self.n}"


class SampleDistinct:
    """Sample distinct operator."""

    def __init__(self, n, of):
        self.n = n
        self.of = of

    def __repr__(self):
        return f"Sample({self.n} of {self.of})"

    def __str__(self):
        return f"| sample {self.n} of {self.of}"


class Distinct:
    def __init__(self, *args):
        self.columns = list(args)

    def _build_column_list(self):
        col_str = ",\n".join([str(col) for col in self.columns])
        return col_str

    def __repr__(self):
        cols = ", ".join([str(col) for col in self.columns])
        return f"Distinct({cols})"

    def __str__(self):

        return f"| distinct {self._build_column_list()}"


class Where:
    def __init__(self, *args):
        self.expressions = list(args)

    def __repr__(self):
        exprs = " and ".join([repr(ex) for ex in self.expressions])
        return f"Where({exprs})"

    def __str__(self):
        exprs = " and ".join([str(ex) for ex in self.expressions])
        return f"| where {exprs}"


class Join:
    def __init__(self, right, on, kind, strategy=None):
        self.right = right
        self.on = [on] if isinstance(on, str) else on
        self.kind = kind
        if strategy in ["broadcast", "shuffle"]:
            self.strategy = strategy
        else:
            self.strategy = None

    def __str__(self):
        on_list = ", ".join([str(col) for col in self.on])
        if self.strategy:
            strategy_str = f"hint.strategy={self.strategy} "
        else:
            strategy_str = ""
        return f"| join kind={self.kind} {strategy_str}(\n\t{self.right}) on {on_list}"


class Summarize:
    def __init__(self, by=None, shuffle=False, shufflekey=None, num_partitions=None, **kwargs):
        # expressions in summarize must be aggregate functions.
        for _, v in kwargs.items():
            assert v.agg
        self.expressions = kwargs
        if by is None:
            self.by = []
        elif isinstance(by, (str, Column)):
            self.by = [by]
        else:
            self.by = by
        # shufflekey takes precedence over shuffle. If shufflekey, then shuffle.
        self.shuffle = bool(shuffle or shufflekey)
        if shufflekey:
            if isinstance(shufflekey, (str, Column)):
                self.shufflekey = [shufflekey]
            else:
                self.shufflekey = shufflekey
        else:
            self.shufflekey = []
        # num_partitions is ignored unless shuffle or shufflekey
        if self.shuffle:
            self.num_partitions = num_partitions

    def __str__(self):
        if self.expressions:
            expr_list = []
            for k, v in self.expressions.items():
                expr_list.append(f"{str(k)}={str(v)}")
            expr_list = ",\n\t".join(expr_list)
            expr_list = f"\n\t{expr_list}"
        else:
            expr_list = ""
        by_list = ", ".join([str(col) for col in self.by])
        if by_list:
            by_list = f"\n\tby {by_list}"
        shuffle_str = ""
        if self.shuffle:
            if self.shufflekey:
                key = ", ".join([str(col) for col in self.shufflekey])
                shuffle_str = f" hint.shufflekey={key}"
            else:
                shuffle_str = " hint.strategy=shuffle"
        partition_str = ""
        if self.shuffle and self.num_partitions:
            partition_str = f" hint.num_partitions={self.num_partitions}"
        clause = f"| summarize{shuffle_str}{partition_str}{expr_list}{by_list}"

        return clause


class Extend:
    def __init__(self, **kwargs):
        self.kwargs = kwargs

    def __str__(self):
        kvs = ",\n\t".join([f"{k}={quote(v)}" for k, v in self.kwargs.items()])
        return f"| extend\n\t{kvs}"


class Property:
    def __init__(self, column, prop):
        self.column = column
        self.prop = prop

    def __str__(self):
        return f"{str(self.column)}.{str(self.prop)}"


class ListLit:
    def __init__(self, *args):
        self.args = list(args)

    def __str__(self):
        in_list = f"({', '.join([quote(arg) for arg in self.args])})"
        return in_list


class Column:
    """A column in a tabular expression."""

    def __init__(self, name: str, dtype: str, ascending: bool = False):
        """"""
        self.name = name
        self.dtype = dtype
        self.ascending = ascending

    def __str__(self):
        return self.name

    def __invert__(self):
        return Prefix(OP.NOT, self, dtype=bool)

    def __eq__(self, rhs):
        return Infix(OP.EQ, self, rhs, dtype=bool)

    def __ne__(self, rhs):
        return Infix(OP.NE, self, rhs, dtype=bool)

    def __lt__(self, rhs):
        return Infix(OP.LT, self, rhs, dtype=bool)

    def __le__(self, rhs):
        return Infix(OP.LE, self, rhs, dtype=bool)

    def __gt__(self, rhs):
        return Infix(OP.GT, self, rhs, dtype=bool)

    def __ge__(self, rhs):
        return Infix(OP.GE, self, rhs, dtype=bool)

    def __add__(self, rhs):
        dtype = int if typeof(self) == typeof(rhs) == int else float
        return Infix(OP.ADD, self, rhs, dtype=dtype)

    def __sub__(self, rhs):
        dtype = int if typeof(self) == typeof(rhs) == int else float
        return Infix(OP.ADD, self, rhs, dtype=dtype)

    def __mul__(self, rhs):
        dtype = int if typeof(self) == typeof(rhs) == int else float
        return Infix(OP.ADD, self, rhs, dtype=dtype)

    def __truediv__(self, rhs):
        dtype = int if typeof(self) == typeof(rhs) == int else float
        return Infix(OP.ADD, self, rhs, dtype=dtype)

    def contains(self, rhs):
        return Infix(OP.CONTAINS, self, rhs, dtype=bool)

    def ncontains(self, rhs):
        return Infix(OP.NCONTAINS, self, rhs, dtype=bool)

    def has(self, rhs):
        return Infix(OP.HAS, self, rhs, dtype=bool)

    def nhas(self, rhs):
        return Infix(OP.NHAS, self, rhs, dtype=bool)

    def isin(self, *args):
        in_list = ListLit(*args)
        return Infix(OP.IN, self, in_list, dtype=bool)

    def sum(self):
        """Aggregate the column by summation."""
        return Prefix(OP.SUM, self, agg=True, dtype=self.dtype)

    def avg(self):
        """Aggregate the column by averaging (arithmetic mean)."""
        return Prefix(OP.AVG, self, agg=True, dtype=float)

    def min(self):
        """Aggregate the column by taking the minimum value."""
        return Prefix(OP.MIN, self, agg=True, dtype=float)

    def max(self):
        """Aggregate the column by taking the maximum value."""
        return Prefix(OP.MAX, self, agg=True, dtype=float)

    def mean(self):
        """Aggregate the column by averaging (arithmetic mean). Alias for avg."""
        return self.avg()

    def percentile(self, *args):
        """Aggregate the column by calculating one or more percentiles."""
        return Prefix(OP.PERCENTILE, self, *args, agg=True, dtype=float)

    def dcount(self, accuracy=1):
        return Prefix(OP.DCOUNT, self, accuracy, agg=True, dtype=int)

    def bag_unpack(self):
        """Expand a dynamic property bag column into one column per property."""
        assert self.dtype in [dict, "dynamic"]
        return Prefix(OP.BAG_UNPACK, self)

    def asc(self):
        """Sort this column in ascending order."""
        self.ascending = True
        return self

    def desc(self):
        """Sort this column in descending order."""
        self.ascending = False
        return self

    def between(self, left, right):
        """Returns True if the column value is between left (inclusive) and right (inclusive)."""
        return Between(self, left, right)

    def nbetween(self, left, right):
        """Returns False if the column value is between left (inclusive) and right (inclusive)."""
        return Between(self, left, right, negate=True)

    def __getattr__(self, attr):
        """Access a field in a dynamic property bag."""
        if self.dtype in [dict, "dynamic"]:
            return Property(self, attr)
        raise AttributeError

    def __repr__(self):
        return f'Column("{self.name}", {self.dtype})'


class Evaluate:
    def __init__(self, expr):
        self.expr = expr

    def __str__(self):
        return f"| evaluate {str(self.expr)}"


def asc(expr):
    """Sort the column in ascending order.

    Parameters
    ----------
    expr:
        A column name string or Column expression instance.
    """
    if isinstance(expr, Column):
        expr.ascending = True
        return expr
    return Column(expr, typeof(expr), ascending=True)


def desc(expr):
    """Sort the column in descending order.

    Parameters
    ----------
    expr:
        A column name string or Column expression instance.
    """
    if isinstance(expr, Column):
        expr.ascending = False
        return expr
    return Column(expr, typeof(expr), asc=False)


class Order:
    def __init__(self, *args):
        self.args = args

    def __str__(self):
        args = []
        for arg in self.args:
            if isinstance(arg, Column):
                if arg.ascending:
                    args.append(quote(arg) + " asc")
                else:
                    args.append(quote(arg))
            else:
                args.append(arg)
        args_str = ",\n\t".join(args)
        return f"| order by\n\t{args_str}"


class Limit:
    def __init__(self, n):
        assert isinstance(n, int)
        self.n = n

    def __str__(self):
        return f"| limit {self.n}"


class Expand:
    def __init__(self, column):
        self.column = column

    def __str__(self):
        return f"| mv-expand {str(self.column)}"


class TableExpr:
    """A table or tabular expression."""

    def __init__(self, name, database, columns=None, ast=None):
        """A tabular expression.

        Parameters
        ----------
        name: str
            The name of the table in the database.
        database: KustoDatabase
            The database containing the table.
        columns: dict or list
            Either:
            1. A dictionary where keys are column names and values are
            data type names, or
            2. A list of Column instances.
        ast: list, default None
            Abstract syntax tree used internally to store a nested list of expressions.
        """
        self._ast = ast or []
        self.name = name
        self.database = database
        if columns is None:
            self.columns = {}
        elif isinstance(columns, (list, tuple)):
            self.columns = {c.name: c for c in columns}
        elif isinstance(columns, dict):
            self.columns = {k: Column(k, v) for k, v in columns.items()}
        else:
            raise ValueError("columns must be a dict or a list of Columns.")

    def __getattr__(self, name):
        try:
            return self.columns[name]
        except KeyError as exc:
            raise AttributeError from exc

    def __getitem__(self, name):
        return self.__getattr__(name)

    def project(self, *args, **kwargs):
        """Project (select) a list of columns.

        Parameters
        ----------
        args: list
            Column names to project.
        kwargs: dict
            Columns to project with renaming, where the key is the new name.
            Right hand side can be a Column or an expression.

        Returns
        -------
        A table expression.
        """
        renamed = {k: Column(k, typeof(v)) for k, v in kwargs.items()}
        columns = {k: v for k, v in self.columns.items() if k in args}
        columns = {**columns, **renamed}
        return TableExpr(
            self.name, self.database, columns=columns, ast=[*self._ast, Project(*args, **kwargs)]
        )

    def collect(self):
        """Compile the expression to a query, execute it, and return results."""
        query_str = str(self)
        return self.database.execute(query_str)

    def count(self):
        """Get the count of rows that would be returned by the expression."""
        return TableExpr(self.name, self.database, columns=self.columns, ast=[*self._ast, Count()])

    def distinct(self, *args):
        """Distinct values in the given column(s)."""
        return TableExpr(
            self.name, self.database, columns=self.columns, ast=[*self._ast, Distinct(*args)]
        )

    def where(self, *args):
        """Filter the expression by one or more predicates."""
        return TableExpr(
            self.name, self.database, columns=self.columns, ast=[*self._ast, Where(*args)]
        )

    def join(self, right, on, kind, *args, strategy=None):
        """Join this table expression to another.

        Parameters
        ----------
        right: TableExpr
            The table to join this table to.
        on: [str]
            The list of columns to join on.
        kind: str
            The kind of join. Options:
            - "inner"
            - "left"
            - "right"
            - "full"
            - "leftsemi"
            - "rightsemi"
            - "leftanti"
            - "rightanti"
        strategy: str, default None
            If "broadcast" then a broadcast join is used.
            If "shuffle" then a shuffle join is used.
            If another value or None, a single-node join strategy is used.
        """
        return TableExpr(
            self.name,
            self.database,
            columns=self.columns,
            ast=[*self._ast, Join(right, on, kind=kind, strategy=strategy)],
        )

    def summarize(self, by=None, shuffle=False, shufflekey=None, num_partitions=None, **kwargs):
        """Aggregate by columns.

        Parameters
        ----------
        by: list, default None
            List of Column instances or column name strings to group by.
        shuffle: bool, default False
            If True, `hint.strategy=shuffle` will be added to the Kusto query.
            The shufflekey parameter takes precedence; if it is not None, then
            shuffle will be ignored.
        shufflekey: [str], str, [Column], Column or bool, default None
            Indicates the key to be used for the shuffle summarize strategy. If
            a string or Column instance, or list thereof, is provided, these
            columns will be used as the shufflekey; `hint.shufflekey=foo, bar`
            will be added to the Kusto query.
        num_partitions: int, default None
            A query hint indicating the number of partitions to be used in the
            shuffle strategy. Has no effect unless `shuffle` or `shufflekey` is
            also provided.
        kwargs: Dict
            Aliased aggregation expressions, e.g. bar=foo.sum()
        """
        # Set table columns to those listed in by, args, kwargs
        if by is None:
            by = []
        elif isinstance(by, (str, Column)):
            by = [by]
        else:
            by = by
        by_cols = {}
        for col in by:
            if isinstance(col, str):
                col = self.columns[col]
            by_cols[col.name] = col

        kwarg_cols = {k: Column(k, typeof(v)) for k, v in kwargs.items()}
        columns = {**by_cols, **kwarg_cols}
        return TableExpr(
            self.name,
            self.database,
            columns=columns,
            ast=[
                *self._ast,
                Summarize(
                    by=by,
                    shuffle=shuffle,
                    shufflekey=shufflekey,
                    num_partitions=num_partitions,
                    **kwargs,
                ),
            ],
        )

    def extend(self, **kwargs):
        """Add new columns calculated from expressions.

        Parameters
        ----------
        kwargs: dict
            Aliased expressions, e.g. foo="bar", baz="quux"
        """
        new_cols = {}
        for key, val in kwargs.items():
            if key not in self.columns:
                new_cols[key] = Column(key, typeof(val))
        columns = {**self.columns, **new_cols}
        new_inst = TableExpr(
            self.name, self.database, columns=columns, ast=[*self._ast, Extend(**kwargs)]
        )
        return new_inst

    def order(self, *args):
        """Order the result set by the given columns.

        Parameters
        ----------
        args: array
            The columns to sort by.
        """
        return TableExpr(
            self.name, self.database, columns=self.columns, ast=[*self._ast, Order(*args)]
        )

    def sort(self, *args):
        """Order the result set by the given columns. Alias for .order().

        Parameters
        ----------
        args: array
            The columns to sort by.
        """
        return self.order(*args)

    def evaluate(self, expr):
        """Evaluate a Kusto plugin expression."""
        return TableExpr(
            self.name, self.database, columns=self.columns, ast=[*self._ast, Evaluate(expr)]
        )

    def limit(self, n):
        """Limit the result set to the first n rows.

        Parameters
        ----------
        n: int
            The number of rows to return.
        """
        return TableExpr(self.name, self.database, columns=self.columns, ast=[*self._ast, Limit(n)])

    def take(self, n):
        """Limit the result set to the first n rows. Alias for .limit().

        Parameters
        ----------
        n: int
            The number of rows to return.
        """
        return self.limit(n)

    def sample(self, n):
        """Randomly sample n rows from the dataset.

        Parameters
        ----------
        n: int
            The number of rows to sample.
        """
        return TableExpr(
            self.name, self.database, columns=self.columns, ast=[*self._ast, Sample(n)]
        )

    def sample_distinct(self, n, column):
        """Randomly sample n rows from the dataset with distinct values in column.

        Parameters
        ----------
        n: int
            The number of rows to sample.
        column:
            The column to sample distinct values from.
        """
        if isinstance(column, str):
            column = self.columns[column]
        return TableExpr(
            self.name,
            self.database,
            columns=self.columns,
            ast=[*self._ast, SampleDistinct(n, column)],
        )

    def mv_expand(self, column):
        """Expand a dynamic column into one row per value.

        Parameters
        ----------
        column: Column
            The column to expand. Must be a dynamic (or dict or array) column.
        """
        if isinstance(column, str):
            column = self.columns[column]
        return TableExpr(
            self.name,
            self.database,
            columns=self.columns,
            ast=[*self._ast, Expand(column)],
        )

    def __str__(self):
        ops = [
            f"cluster('{self.database.cluster}').database('{self.database.database}').['{self.name}']",
            *self._ast,
        ]
        query_str = "\n".join([str(op) for op in ops]) + "\n"
        return query_str


__all__ = ["TableExpr"]



from kusto_tool.database import KustoDatabase
from kusto_tool.expression import Limit, TableExpr


def test_limit():
    """limit prints."""
    assert str(Limit(1000)) == "| limit 1000"


def test_limit_tbl():
    """limit works on a tbl."""
    assert (
        str(TableExpr("tbl", KustoDatabase("c", "db")).limit(1000))
        == "cluster('c').database('db').['tbl']\n| limit 1000\n"
    )


def test_take():
    """take works."""
    assert (
        str(TableExpr("tbl", KustoDatabase("c", "db")).take(1000))
        == "cluster('c').database('db').['tbl']\n| limit 1000\n"
    )


def test_sample():
    """sample works."""
    assert (
        str(TableExpr("tbl", KustoDatabase("c", "db")).sample(1000))
        == "cluster('c').database('db').['tbl']\n| sample 1000\n"
    )


def test_sample_distinct():
    """sample-distinct works."""
    assert (
        str(
            TableExpr("tbl", KustoDatabase("c", "db"), columns=dict(foo=str)).sample_distinct(
                1000, "foo"
            )
        )
        == "cluster('c').database('db').['tbl']\n| sample 1000 of foo\n"
    )



from kusto_tool.database import KustoDatabase
from kusto_tool.expression import Column, Order, TableExpr, asc


def test_order_1arg():
    """order by works"""
    assert str(Order(Column("foo", str))) == "| order by\n\tfoo"


def test_order_2args():
    """order by works with 2 args"""
    assert str(Order(Column("foo", str), Column("bar", str))) == "| order by\n\tfoo,\n\tbar"


def test_order_asc():
    """order works with asc"""
    col = Column("foo", str).asc()
    expr = Order(col)
    assert str(expr) == "| order by\n\tfoo asc"


def test_order_asc_multiple():
    """order works with asc"""
    expr = Order(Column("foo", str).desc(), Column("bar", str).asc(), Column("baz", str).desc())
    assert str(expr) == "| order by\n\tfoo,\n\tbar asc,\n\tbaz"


def test_order_tbl():
    """order by works with a table"""
    tbl = TableExpr("tbl", KustoDatabase("cluster", "db"), columns={"foo": str})
    assert (
        str(tbl.order(tbl.foo)) == "cluster('cluster').database('db').['tbl']\n| order by\n\tfoo\n"
    )


def test_order_tbl_asc():
    """order by works with a table"""
    tbl = TableExpr("tbl", KustoDatabase("cluster", "db"), columns={"foo": str})
    assert (
        str(tbl.order(tbl.foo.asc()))
        == "cluster('cluster').database('db').['tbl']\n| order by\n\tfoo asc\n"
    )


def test_sort_tbl():
    """sort alias works works with a table"""
    tbl = TableExpr("tbl", KustoDatabase("cluster", "db"), columns={"foo": str})
    assert (
        str(tbl.sort(tbl.foo)) == "cluster('cluster').database('db').['tbl']\n| order by\n\tfoo\n"
    )


def test_order_str():
    """order works with a string column name"""
    assert str(Order("foo")) == "| order by\n\tfoo"


def test_order_str_asc():
    """order works with a string column name"""
    assert str(Order(asc("foo"))) == "| order by\n\tfoo asc"



from kusto_tool.expression import Column


def test_column_contains():
    col = Column("foo", str)
    assert str(col.contains("bar")) == "foo contains 'bar'"


def test_column_has():
    col = Column("foo", str)
    assert str(col.has("bar")) == "foo has 'bar'"


def test_column_ncontains():
    assert str(Column("foo", str).ncontains("bar")) == "foo !contains 'bar'"


def test_column_nhas():
    assert str(Column("foo", str).nhas("bar")) == "foo !has 'bar'"


def test_between():
    assert str(Column("foo", int).between(1, 3)) == "foo between(1 .. 3)"


def test_nbetween():
    assert str(Column("foo", int).nbetween(1, 3)) == "foo !between(1 .. 3)"



import os
import shutil
from pathlib import Path

import nox

nox.options.reuse_existing_virtualenvs = True


@nox.session(python=["3.7", "3.8", "3.9", "3.10"])
def test(session):
    """Run the tests."""
    session.install(".")
    session.install("pytest")
    session.run("pytest")


@nox.session(python="3.10")
def coverage(session) -> None:
    """Upload coverage data."""
    session.install(".")
    session.install("coverage[toml]")
    session.run("coverage", "report", "--fail-under=0")


@nox.session(python="3.10")
def docs(session):
    """Build the documentation."""
    session.install(".")
    session.run("sphinx-build", "docs", "docs/_build")

    # args = session.posargs or ["-W", "-n", "docs", "docs/_build"]

    # if session.interactive and not session.posargs:
    #     args = ["-a", "--watch=docs/_static", "--open-browser", *args]

    # builddir = Path("docs", "_build")
    # if builddir.exists():
    #     shutil.rmtree(builddir)

    # session.install("-r", "docs/requirements.txt")

    # if session.interactive:
    #     session.run("sphinx-autobuild", *args, external=True)
    # else:
    #     session.run("sphinx-build", *args, external=True)



from kusto_tool.expression import Column, Summarize, TableExpr
from kusto_tool.function import strcat, sum
from pytest import raises

from .fake_database import FakeDatabase


def test_summarize_by_list_str():
    """by clause can be included in summarize as a list of strings"""
    result = str(Summarize(by=["foo", "bar"]))
    expected = "| summarize\n\tby foo, bar"
    assert result == expected


def test_summarize_by_str():
    """by clause can be included in summarize as a single string"""
    result = str(Summarize(by="foo"))
    expected = "| summarize\n\tby foo"
    assert result == expected


def test_summarize_by_list_col():
    """by clause can be included in summarize as a list of columns"""
    result = str(Summarize(by=[Column("foo", str), Column("bar", str)]))
    expected = "| summarize\n\tby foo, bar"
    assert result == expected


def test_summarize_by_col():
    """by clause can be included in summarize as a single Column"""
    result = str(Summarize(by=Column("foo", str)))
    expected = "| summarize\n\tby foo"
    assert result == expected


def test_summarize_noby():
    """by clause can be omitted from summarize"""
    result = str(Summarize(baz=Column("bar", int).sum()))
    expected = "| summarize\n\tbaz=sum(bar)"
    assert result == expected


def test_summarize_sum_by():
    """summarize can use sum()"""
    result = str(Summarize(baz=Column("bar", int).sum(), by="foo"))
    expected = "| summarize\n\tbaz=sum(bar)\n\tby foo"
    assert result == expected


def test_summarize_sum_two_cols():
    """Summarize can work with two columns"""
    result = str(
        Summarize(baz=Column("bar", int).sum(), quux=Column("bar", int).dcount(), by="foo")
    )
    expected = "| summarize\n\tbaz=sum(bar),\n\tquux=dcount(bar, 1)\n\tby foo"
    assert result == expected


def test_summarize_dcount_by():
    """summarize can use sum()"""
    result = str(Summarize(baz=Column("bar", int).dcount(), by="foo"))
    expected = "| summarize\n\tbaz=dcount(bar, 1)\n\tby foo"
    assert result == expected


def test_summarize_strategy():
    """shuffle parameter works"""
    result = str(Summarize(baz=Column("bar", int).sum(), by="foo", shuffle=True))
    expected = "| summarize hint.strategy=shuffle\n\tbaz=sum(bar)\n\tby foo"
    assert result == expected


def test_summarize_strategy_shufflekey_list_col():
    """shufflekey parameter works for list of columns"""
    result = str(
        Summarize(
            baz=Column("bar", int).sum(),
            by="foo",
            shufflekey=[Column("baz", str), Column("quux", str)],
        )
    )
    expected = "| summarize hint.shufflekey=baz, quux\n\tbaz=sum(bar)\n\tby foo"
    assert result == expected


def test_summarize_strategy_shufflekey_list_str():
    """shufflekey parameter works for list of strings"""
    result = str(Summarize(baz=Column("bar", int).sum(), by="foo", shufflekey=["baz", "quux"]))
    expected = "| summarize hint.shufflekey=baz, quux\n\tbaz=sum(bar)\n\tby foo"
    assert result == expected


def test_summarize_strategy_shufflekey_col():
    """shufflekey parameter works for as single column"""
    result = str(Summarize(baz=Column("bar", int).sum(), by="foo", shufflekey=Column("baz", str)))
    expected = "| summarize hint.shufflekey=baz\n\tbaz=sum(bar)\n\tby foo"
    assert result == expected


def test_summarize_strategy_shufflekey_str():
    """shufflekey parameter works for a single string"""
    result = str(Summarize(baz=Column("bar", int).sum(), by="foo", shufflekey="baz"))
    expected = "| summarize hint.shufflekey=baz\n\tbaz=sum(bar)\n\tby foo"
    assert result == expected


def test_summarize_strategy_partitions():
    """num_partitions parameter works"""
    result = str(
        Summarize(
            baz=Column("bar", int).sum(),
            by="foo",
            shufflekey=[Column("baz", str), Column("quux", str)],
            num_partitions=10,
        )
    )
    expected = (
        "| summarize hint.shufflekey=baz, quux hint.num_partitions=10\n\tbaz=sum(bar)\n\tby foo"
    )
    assert result == expected


def test_summarize_strategy_partitions_ignored():
    """num_partitions is ignored if not(shuffle or shufflekey)"""
    result = str(
        Summarize(
            baz=Column("bar", int).sum(),
            by="foo",
            num_partitions=10,
        )
    )
    expected = "| summarize\n\tbaz=sum(bar)\n\tby foo"
    assert result == expected


def test_tableexpr_summarize():
    """TableExpr.summarize calls Summarize correctly"""
    db = FakeDatabase("help", "Samples")
    tbl = TableExpr("tbl", database=db, columns={"foo": str, "bar": int})
    query = tbl.summarize(sum_foo=Column("foo", str).sum(), by="bar")
    assert "sum_foo" in query.columns
    assert "bar" in query.columns
    assert "foo" not in query.columns
    result = str(query)
    expected = """cluster('help').database('Samples').['tbl']
| summarize
\tsum_foo=sum(foo)
\tby bar
"""
    assert result == expected


def test_tableexpr_summarize_noby():
    """TableExpr.summarize calls Summarize correctly without a by clause"""
    db = FakeDatabase("help", "Samples")
    tbl = TableExpr("tbl", database=db, columns={"foo": str, "bar": int})
    query = tbl.summarize(sum_foo=Column("foo", str).sum())
    assert "sum_foo" in query.columns
    assert "bar" not in query.columns
    assert "foo" not in query.columns
    result = str(query)
    expected = """cluster('help').database('Samples').['tbl']
| summarize
\tsum_foo=sum(foo)
"""
    assert result == expected


def test_tableexpr_summarize_function():
    """TableExpr.summarize calls Summarize correctly with the sum() function."""
    db = FakeDatabase("help", "Samples")
    tbl = TableExpr("tbl", database=db, columns={"foo": str, "bar": int})
    query = tbl.summarize(sum_foo=sum(Column("foo", str)))
    assert "sum_foo" in query.columns
    assert "bar" not in query.columns
    assert "foo" not in query.columns
    result = str(query)
    expected = """cluster('help').database('Samples').['tbl']
| summarize
\tsum_foo=sum(foo)
"""
    assert result == expected


def test_tableexpr_summarize_function_str():
    """TableExpr.summarize calls Summarize correctly for a string arg."""
    db = FakeDatabase("help", "Samples")
    tbl = TableExpr("tbl", database=db, columns={"foo": str, "bar": int})
    query = tbl.summarize(sum_foo=sum("foo"))
    assert "sum_foo" in query.columns
    assert "bar" not in query.columns
    assert "foo" not in query.columns
    result = str(query)
    expected = """cluster('help').database('Samples').['tbl']
| summarize
\tsum_foo=sum(foo)
"""
    assert result == expected


def test_tableexpr_summarize_function_missing_col():
    """TableExpr.summarize calls Summarize correctly"""
    db = FakeDatabase("help", "Samples")
    tbl = TableExpr("tbl", database=db, columns={"foo": str, "bar": int})
    with raises(AttributeError):
        tbl.summarize(sum_foo=tbl.baz.sum())


def test_tableexpr_summarize_function_str_missing_col():
    """TableExpr.summarize calls Summarize correctly for a nonexisting string arg.
    Note: This query will fail on the server side since we cannot tell if "baz"
    is a real column in the dataset.
    """
    db = FakeDatabase("help", "Samples")
    tbl = TableExpr("tbl", database=db, columns={"foo": str, "bar": int})
    query = tbl.summarize(sum_baz=sum("baz"))
    assert "sum_baz" in query.columns
    assert "bar" not in query.columns
    assert "foo" not in query.columns
    result = str(query)
    expected = """cluster('help').database('Samples').['tbl']
| summarize
\tsum_baz=sum(baz)
"""
    assert result == expected


def test_summarize_nonagg_fails():
    db = FakeDatabase("help", "Samples")
    tbl = TableExpr("tbl", database=db, columns={"foo": str, "bar": int})
    with raises(AssertionError):
        tbl.summarize(sum_foo=strcat(Column("foo", str)))



"""A high-level API for querying Kusto."""
__version__ = "0.3.1"

from .database import Cluster, KustoDatabase, cluster
from .expression import TableExpr

__all__ = ["KustoDatabase", "Cluster", "cluster", "TableExpr"]
